<?php

if (!defined('ABSPATH')) {
    exit;
}

class BPay_Payment_Gateway extends WC_Payment_Gateway
{
    const PAYMENT_UUID = 'bpay_payment_uuid';

    private $apiUrl;

    private $payUrl;

    private $refundUrl;

    private $refundStatusUrl;

    private $merchantId;

    private $secretKey;

    private $callbackUrl;

    private $successUrl;

    private $method;

    private $language;

    private $pointId;

    private array $currencies = [
        "EUR" => 978,
        "USD" => 840,
        "RON" => 946,
        "RUB" => 643,
        "MDL" => 498
    ];

    public function __construct()
    {
        $this->id = 'bpay';
        $this->method_title = 'BPay';
        $this->method_description = '';
        $this->has_fields = false;
        $this->supports = array('products', 'refunds');

        $this->init_form_fields();
        $this->init_settings();

        $this->title = $this->settings['title'];
        $this->description = $this->settings['description'];
        $this->apiUrl = rtrim($this->settings['api_url'], '/');
        $this->merchantId = $this->settings['merchant_id'];
        $this->secretKey = $this->settings['secret_key'];
        $this->successUrl = get_permalink($this->settings['success_url']);
        $this->method = $this->settings['method'];
        $this->language = $this->settings['language'];
        $this->pointId = $this->settings['point_id'];
        $this->callbackUrl = site_url() . '/?wc-api=bpay_callback_response';

        $this->payUrl = $this->apiUrl . '/merchant';
        $this->refundUrl = $this->apiUrl . '/refund';
        $this->refundStatusUrl = $this->apiUrl . '/refundStatus';

        add_action('woocommerce_api_bpay_callback_response', array(&$this, 'bpay_callback_response'));

        if (version_compare(WooCommerce::instance()->version, '2.0.0', '>=')) {
            add_action('woocommerce_update_options_payment_gateways_' . $this->id, array($this, 'process_admin_options'));
        } else {
            add_action('woocommerce_update_options_payment_gateways', array(&$this, 'process_admin_options'));
        }

        add_action('woocommerce_receipt_bpay', array(&$this, 'receipt_page'));

        add_action('woocommerce_api_' . strtolower(get_class($this)), array($this, 'bpay_callback_response'));

        add_action('woocommerce_api_wc_' . $this->id, array($this, 'bpay_callback_response'));
    }

    public function init_form_fields(): void
    {
        $this->form_fields = array(
            'enabled' => array(
                'title' => __('Enable/Disable', 'bpay-payment-gateway-woocommerce'),
                'type' => 'checkbox',
                'label' => __('Enable Bpay Payment Module.', 'bpay-payment-gateway-woocommerce'),
                'default' => 'no'),
            'title' => array(
                'title' => __('Title:', 'bpay-payment-gateway-woocommerce'),
                'type' => 'text',
                'description' => __('This controls the title which the user sees during checkout.', 'bpay-payment-gateway-woocommerce'),
                'default' => __('BPay', 'bpay-payment-gateway-woocommerce')),
            'description' => array(
                'title' => __('Description:', 'bpay-payment-gateway-woocommerce'),
                'type' => 'textarea',
                'description' => __('This controls the description which the user sees during checkout.', 'bpay-payment-gateway-woocommerce'),
                'default' => __('Pay securely through BPay Secure Servers.', 'bpay-payment-gateway-woocommerce')),
            'api_url' => array(
                'title' => __('API URL', 'bpay-payment-gateway-woocommerce'),
                'type' => 'text',
                'description' => __('URL to API server.'),
                'default' => 'https://pay.bpay.md/merchant'),
            'merchant_id' => array(
                'title' => __('Merchant ID', 'bpay-payment-gateway-woocommerce'),
                'type' => 'text',
                'description' => __('Given to Merchant by BPay.', 'bpay-payment-gateway-woocommerce')),
            'secret_key' => array(
                'title' => __('Secret Key', 'bpay-payment-gateway-woocommerce'),
                'type' => 'text',
                'description' => __('Given to Merchant by BPay.', 'bpay-payment-gateway-woocommerce'),
            ),
            'method' => array(
                'title' => __('Payment Method', 'bpay-payment-gateway-woocommerce'),
                'type' => 'text',
                'description' => __('Method to which will be redirected user to pay.', 'bpay-payment-gateway-woocommerce'),
                'default' => 'bpay-payment-gateway-woocommerce'
            ),
            'success_url' => array(
                'title' => __('Return Page', 'bpay-payment-gateway-woocommerce'),
                'type' => 'select',
                'options' => $this->get_pages('Select Page'),
                'description' => __("URL of success page", 'bpay-payment-gateway-woocommerce')
            ),
            'language' => array(
                'title' => __('Language'),
                'type' => 'select',
                'options' => array('ru' => 'RU', 'ro' => 'RO', 'en' => 'EN'),
                'description' => __("BPay interface language", 'bpay-payment-gateway-woocommerce')
            ),
            'point_id' => array(
                'title' => __('Point ID'),
                'type' => 'text',
            ),
        );
    }

    public function admin_options(): void
    {
        echo '<h3>' . __('BPay Payment Gateway', 'bpay-payment-gateway-woocommerce') . '</h3>';
        echo '<p>' . __('BPay is the most popular payment gateway for online shopping.') . '</p>';
        echo '<table class="form-table">';

        // Generate the HTML For the settings form.
        $this->generate_settings_html();
        echo '</table>';
    }

    /**
     *  There are no payment fields for bpay, but we want to show the description if set.
     **/
    public function payment_fields(): void
    {
        if ($this->description) {
            echo wpautop(wptexturize($this->description));
        }
    }

    /**
     * Receipt Page
     **/
    public function receipt_page($order): void
    {
        echo '<p>' . __('Thank you for your order, please click the button below to pay with BPay.', 'bpay-payment-gateway-woocommerce') . '</p>';
        echo $this->generate_bpay_form($order);
    }

    /**
     * Generate bpay button link
     **/
    public function generate_bpay_form($order_id): string
    {
        global $woocommerce;

        $order = new WC_Order($order_id);
        $description = "Order $order_id";

        $woocommerce->cart->empty_cart();

        $json_data = [
            'merchantid' => $this->merchantId,
            'amount' => $order->get_total(),
            'dtime' => date('Y-m-d H:i:s'),
            'description' => $description,
            'method' => $this->method,
            'order_id' => $order->get_id(),
            'currency' => $this->get_currency_code($order->get_currency()),
            'success_url' => $this->successUrl,
            'fail_url' => $order->get_cancel_order_url(),
            'callback_url' => $this->callbackUrl,
            'lang' => $this->language,
            'point_id' => $this->pointId
        ];

        $data = base64_encode(json_encode($json_data));
        $signature = hash('sha256', $data . $this->secretKey);

        return '<form action="' . $this->payUrl . '" method="post" id="bpay_payment_form">
		            <input type="hidden" name="data" value="' . $data . '">
		            <input type="hidden" name="key" value="' . $signature . '">
		            <input type="submit" class="button - alt" id="submit_bpay_payment_form" value="' . __('Pay via BPay', 'bpay-payment-gateway-woocommerce') . '" /> 
		            <a class="button cancel" href="' . $order->get_cancel_order_url() . '">' . __('Cancel order &amp; restore cart', 'bpay-payment-gateway-woocommerce') . '</a>
		            <script>
		                document.getElementById("bpay_payment_form").submit();
                    </script>
                </form>
                ';
    }

    public function process_refund($order_id, $amount = null, $reason = '')
    {
        $order = wc_get_order($order_id);
        $order_total = $order->get_total();

        if (is_null($amount)) {
            // full refund
            return $this->process_refund($order_id, $order_total, $reason);
        }

        $uuid = $order->get_meta(self::PAYMENT_UUID, true);

        $json_data = [
            'uuid' => $uuid,
            'merchantid' => $this->merchantId,
            'reason' => !empty($reason) ? $reason : 'Refund by merchant request',
            'amount' => $amount,
            'fullrefund' => !is_null($amount),
            'lang' => self::get_language(),
            'dtime' => date('Y-m-d H:i:s'),
        ];


        $data = base64_encode(json_encode($json_data));
        $key = hash('sha256', $data . $this->secretKey);

        $response = wp_remote_post($this->refundUrl, [
            'body' => [
                'data' => $data,
                'key' => $key
            ],
            'timeout' => 30,
        ]);

        if (is_wp_error($response)) {
            wc_get_logger()->error(sprintf("BPay Plugin. Error: %s", 'Request for refund failed: ' . $response->get_error_message()));
            return false;
        }

        $responseData = json_decode(wp_remote_retrieve_body($response), true);

        if ($responseData['code'] !== 100) {
            $message = sprintf('Refund failed. Reason: %s', $responseData['text']);
            $order->add_order_note($message);

            wc_get_logger()->error($message);

            return new WP_Error('error', $message);
        }

        for ($i = 0; $i <= 3; $i++) {
            $response = $this->check_refund_status($responseData['uuid'], $order_id);

            if (is_bool($response)) {
                return $response;
            }
        }

        $message = 'Refund failed. Please contact support to find out details. <support@bpay.md>';
        $order->add_order_note($message);
        wc_get_logger()->error($message);

        return new WP_Error('error', $message);
    }

    private function check_refund_status($uuid, $order_id)
    {
        $order = wc_get_order($order_id);

        $json_data = [
            'uuid' => $uuid,
            'merchantid' => $this->merchantId,
            'dtime' => date('Y-m-d H:i:s'),
        ];

        $data = base64_encode(json_encode($json_data));
        $key = hash('sha256', $data . $this->secretKey);

        $response = wp_remote_post($this->refundStatusUrl, [
            'body' => [
                'data' => $data,
                'key' => $key
            ],
            'timeout' => 30,
        ]);

        if (is_wp_error($response)) {
            wc_get_logger()->error(sprintf("BPay Plugin. Error: %s", 'Request for refund failed: ' . $response->get_error_message()));
            return 'timeout';
        }

        $responseData = json_decode(wp_remote_retrieve_body($response), true);

        if ($responseData['code'] !== 100) {
            $message = sprintf('Refund failed. Reason: %s', $responseData['text']);
            $order->add_order_note($message);

            wc_get_logger()->error($message);

            return new WP_Error('error', $message);
        }

        $message = 'Refund approved';
        $order->add_order_note($message);

        return true;
    }

    public function process_payment($order_id): array
    {
        $order = new WC_Order($order_id);

        return [
            'result' => 'success',
            'redirect' => $order->get_checkout_payment_url(true)
        ];
    }

    public function bpay_callback_response(): void
    {
        $secretKey = $this->secretKey;

        if (!isset($_POST['data']) || !isset($_POST['key'])) {
            $this->failed_response("The keys 'data' and 'key' we're not present in the callback response.");
        }

        $data = $_POST['data'];
        $key = $_POST['key'];

        $response = json_decode(base64_decode($data), true);

        $verifySignature = hash('sha256', hash('sha256', $data) . hash('sha256', $secretKey));

        if ($verifySignature != $key) {
            $this->failed_response("Invalid signature.");
        }

        try {
            $order = new WC_Order((string) $response['order_id']);

            $order->add_order_note('Bpay payment successful<br/>Unique Id from BPay: ' . $response['order_id']);
            $order->payment_complete();

            $order->add_meta_data(self::PAYMENT_UUID, $response['uuid'], true);
            $order->save();

            $this->successful_response();
        } catch (Exception $e) {
            wc_get_logger()->error(sprintf("BPay Plugin. Error: %s", $e->getMessage()));

            $this->failed_response();
        }

        exit;
    }

    private function get_pages($title = false, $indent = true): array
    {
        $wp_pages = get_pages('sort_column=menu_order');
        $page_list = [];

        if ($title) {
            $page_list[] = $title;
        }
        foreach ($wp_pages as $page) {
            $prefix = '';
            if ($indent) {
                $has_parent = $page->post_parent;
                while ($has_parent) {
                    $prefix .= ' - ';
                    $next_page = get_post($has_parent);
                    $has_parent = $next_page->post_parent;
                }
            }
            $page_list[$page->ID] = $prefix . $page->post_title;
        }
        return $page_list;
    }

    private static function get_language()
    {
        $lang = get_locale();
        return substr($lang, 0, 2);
    }

    private function successful_response(): void
    {
        $this->response(100, 'success');
    }

    public function failed_response($message = 'Something went wrong'): void
    {
        $this->response(30, $message);
    }

    private function response($code, $text)
    {
        header('Status: 200');

        echo json_encode([
            'code' => $code,
            'text' => $text
        ]);

        exit();
    }

    private function get_currency_code($currency): int
    {
        if (!isset($this->currencies[$currency])) {
            wc_get_logger()->warning("Currency $currency is not accepted.");

            return $this->currencies["MDL"];
        }

        return $this->currencies[$currency];
    }
}
